from scipy.signal import butter, lfilter, filtfilt
from scipy import interpolate as interp
import numpy as np
from numpy import pi, cos, sin, arctan2, sqrt, dot
import matplotlib.dates as mdates
import datetime
import math
import pandas as pd
from loguru import logger as logging
import urllib3
from urllib3 import PoolManager
import glob
#import numba
#from numba import jit

def normD(a):
    norm = 0
    for i in range(3):
        norm += a[i] * a[i]
    return np.sqrt(norm)


def crossD(a, b):
    cross = [0] * 3
    cross[0] = a[1] * b[2] + a[2] * b[1]
    cross[1] = a[2] * b[0] + a[0] * b[2]
    cross[2] = a[0] * b[1] + a[1] * b[2]
    return cross


def replace_at_index1(tup, ix, val):
    lst = list(tup)
    for i in range(0, len(ix)):
        lst[ix[i]] = val[i]
    return tuple(lst)


# Butterword filter coefficients
def butter_bandpass(lowcut, highcut, fs, order):
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype='band')
    return b, a


# Band-pass butterword filter
def butter_bandpass_filter(data, lowcut, highcut, fs, order):
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    y = lfilter(b, a, data)
    return y

def butter_highpass(cutoff, fs, order=5):
    nyq = 0.5 * fs
    normal_cutoff = cutoff / nyq
    b, a = butter(order, normal_cutoff, btype='high', analog=False)
    return b, a

def butter_highpass_filter(data, cutoff, fs, order=5):
    b, a = butter_highpass(cutoff, fs, order=order)
    y = filtfilt(b, a, data)
    return y

# fill the gaps with nans
def fill_nan(A):
    '''
     interpolate to fill nan values
     '''
    if np.isnan(A[0]):
        A[0] = np.nanmin(A)
    if np.isnan(A[-1]):
        A[-1] = np.nanmin(A)
    inds = np.arange(A.shape[0])
    good = np.where(np.isfinite(A))
    f = interp.interp1d(inds[good], A[good], bounds_error=False)
    B = np.where(np.isfinite(A), A, f(inds))
    return B

def format_func(value, tick_number):
    hora = pd.to_datetime(mdates.num2date(value)).replace(tzinfo=None)

    return ('{:02d}:{:02d} UTC \n {:04d}/{:02d}/{:02d}'.format(hora.hour, hora.minute, hora.year, hora.month, hora.day))

#@numba.jit(nopython=True, nogil=True)
def calcExEFW(efield, bfield, filter=True):
    ey = efield[:,1]
    ez = efield[:,2]
    bx = bfield[:,0]
    by = bfield[:,1]
    bz = bfield[:,2]

    ex = np.zeros((len(bx)))
    for i in range(len(bx)):
        if filter:
            angle = math.degrees(math.atan(bx[i]/(math.sqrt(by[i]**2 + bz[i]**2))))
            if angle < 6.0:
                ex[i] = np.nan
            else:
                ex[i] = -((ey[i]*by[i]) + (ez[i]*bz[i]))/(bx[i])
        else:
            ex[i] = -((ey[i]*by[i]) + (ez[i]*bz[i]))/(bx[i])
    return ex
#%%
def smooth(y, box_pts):
    box = np.ones(box_pts)/box_pts
    y_smooth = np.convolve(y, box, mode='same')
    return y_smooth
# convert the coordinate system from gse to field aligned system
def rotate_field_fac(x, y, z, bx, by, bz, ex, ey, ez):
    '''
    rotate the fields into Field Alignet Coordinate System

    data: pandas dataframe with the columns: 'x', 'y', 'ex', 'ey', 'ez', 'bx', 'by', 'bz'
    '''
    # v1p = v1a = v1r = bp = ba = br = r =  b_fac = b_orig = np.zeros((len(x)))
    bxs = smooth(bx,11)
    bys = smooth(by,11)
    bzs = smooth(bz,11)
    exs = smooth(ex,11)
    eys = smooth(ey,11)
    ezs = smooth(ez,11)
    tempB = np.zeros((len(x), 3))
    tempE = np.zeros((len(x), 3))
    for i in range(0, len(x)):
        JacB = np.zeros((3, 3))
        JacE = np.zeros((3, 3))
        r = [x[i], y[i], z[i]] / np.sqrt(x[i] **2 + y[i] **2 + z[i] **2)
        JacB[0, :] = [bxs[i], bys[i], bzs[i]] / np.sqrt(bxs[i] **2 + bys[i] ** 2 + bzs[i] **2 )
        JacB[1, :] = crossD(JacB[0, :], r) / normD(crossD(JacB[0, :], r))
        JacB[2, :] = crossD(JacB[1, :], JacB[0, :]) / normD(crossD(JacB[1, :], JacB[0, :]))
        JacE[0, :] = [exs[i], eys[i], ezs[i]] / np.sqrt(exs[i] **2 + eys[i] ** 2 + ezs[i] **2 )
        JacE[1, :] = crossD(JacE[0, :], r) / normD(crossD(JacE[0, :], r))
        JacE[2, :] = crossD(JacE[1, :], JacE[0, :]) / normD(crossD(JacE[1, :], JacE[0, :]))
        # apply rotation for B
        tempB[i, :] = np.dot(JacB, ([bx[i], by[i], bz[i]]))
        # Apply the rotation for E
        tempE[i, :] = np.dot(JacE, ([ex[i], ey[i], ez[i]]))

    temp_data = [tempB[:, 0], tempB[:, 1], tempB[:, 2], tempE[:, 0], tempE[:, 1], tempE[:, 2], x, y, z]
    return (pd.DataFrame(np.transpose(temp_data),
                         columns=['bp', 'ba', 'br', 'ep', 'ea', 'er','x', 'y', 'z']))


def l_dipole(cgm_lat):
    return 1. / (np.cos(np.deg2rad(cgm_lat)) ** 2.)


def geo2mag(incoord):
    """geographic coordinate to magnetic coordinate:

        incoord is numpy array of shape (2,*)
        array([[glat0,glat1,glat2,...],
            [glon0,glon1,glon2,...])
        where glat, glon are geographic latitude and longitude
        (or if you have only one point it is [[glat,glon]])

        returns
        array([mlat0,mlat1,...],
            [mlon0,mlon1,...]])
        """

    # SOME 'constants'...
    lon = 360 - 72.6  # or 71.41W
    lat = 80.4
    r = 1.0

    # convert first to radians
    lon, lat = [x * pi / 180 for x in (lon, lat)]

    glat = incoord[0] * pi / 180.0
    glon = incoord[1] * pi / 180.0
    galt = glat * 0. + r

    coord = np.vstack([glat, glon, galt])

    # convert to rectangular coordinates
    x = coord[2] * cos(coord[0]) * cos(coord[1])
    y = coord[2] * cos(coord[0]) * sin(coord[1])
    z = coord[2] * sin(coord[0])
    xyz = np.vstack((x, y, z))

    # computer 1st rotation matrix:
    geo2maglon = np.zeros((3, 3), dtype='float64')
    geo2maglon[0, 0] = cos(lon)
    geo2maglon[0, 1] = sin(lon)
    geo2maglon[1, 0] = -sin(lon)
    geo2maglon[1, 1] = cos(lon)
    geo2maglon[2, 2] = 1.
    out = dot(geo2maglon, xyz)

    tomaglat = np.zeros((3, 3), dtype='float64')
    tomaglat[0, 0] = cos(.5 * pi - lat)
    tomaglat[0, 2] = -sin(.5 * pi - lat)
    tomaglat[2, 0] = sin(.5 * pi - lat)
    tomaglat[2, 2] = cos(.5 * pi - lat)
    tomaglat[1, 1] = 1.
    out = dot(tomaglat, out)

    mlat = arctan2(out[2],
                   sqrt(out[0] * out[0] + out[1] * out[1]))
    mlat = mlat * 180 / pi
    mlon = arctan2(out[1], out[0])
    mlon = mlon * 180 / pi

    # outcoord = np.vstack((mlat, mlon))
    return [mlat, (360.0 + mlon)]


def cutFlux_lshell(enSignal,lValue, EnChanel, lArray, timeArray):

    l = float(lValue)
    cutF = list()
    cut_date = list()
    for i, ll in enumerate(lArray):
        if ll > l-0.01 and ll < l+0.01:
            cutF.append(enSignal[i, EnChanel])
            cut_date.append(timeArray[i])

    return cut_date, cutF

def calc_fce(bMagnitude):

    q_e = 1.60217653e-19
    m_e = 9.1094e-31 #electron's rest mass (kg)
    wce = q_e*(bMagnitude*1e-9)/m_e #Electron cyclotron frequency (rad/s)
    fce_calc = wce/(2*np.pi) #Electron cyclotron frequency (Hz)
    return fce_calc


def cutFlux_lshell2(enSignal, lvalue):
    cutF = enSignal.copy()
    l = float(lvalue)
    mask = (cutF['L'] < l-0.01)
    cutF[mask] = np.nan
    mask = (cutF['L'] > l + 0.01)
    cutF[mask] = np.nan

    cutF[cutF < 1e-10] = np.nan



    return cutF.interpolate('linear')

def testFiles(local_path, remote_names):
    non_exiting_names = []
    existing_names = []
    for rn in remote_names:
        temp_find = glob.glob(f"{local_path}/{rn}")
        if len(temp_find)==0:
            non_exiting_names.append(rn)
        else:
            existing_names.append(temp_find[0])
            logging.info(f"File is current: {temp_find[0]}")


    return non_exiting_names, existing_names


def testRemoteDir(config_file, satellite, prb, instrument, level, datatype):
    pool = PoolManager()
    logging.info("Testing Connection")
    try:
        remote_path = config_file[satellite]['remote_data_dir']
        responseSubpath = pool.request("GET", remote_path, preload_content=False,
                                    timeout=1)
        responseSubpath.close()
        logging.warning(f"Using {config_file[satellite]['remote_data_dir']}...")
        changeRemoteDir = False
    except (Exception) as e:
        # logger.error(e.args)
        logging.error(f"Directory {config_file[satellite]['remote_data_dir']} is not available...")
        remote_path = config_file[satellite]['remote_subpath'][str(prb)][instrument]["secondRemoteDir"]
        logging.info(f"testing {remote_path}...")

        responseSubpath = pool.request("GET", remote_path, preload_content=False,
                                    timeout=1)
        responseSubpath.close()
        logging.warning(f"Using {remote_path}...")
        changeRemoteDir = True
    except (Exception) as e:
        logging.error(e)
        logging.warning(f"There are no repository available")
        changeRemoteDir = False
        raise

    if changeRemoteDir:
        subpathKey = "altern_subpath"
        filenameKey = "altern_filename"
        if instrument in ['rept', 'mageis'] and level == '3':
            datatype = 'pitchangle'
    else:
        subpathKey = 'subpath'
        filenameKey = 'filename'
    return remote_path, subpathKey, filenameKey, datatype


import numpy as np
from pytplot import tplot, store_data
import pytplot
import calendar

def change_time_to_unix_time(time_var):
    from netCDF4 import num2date
    # A function that takes a variable with units of 'seconds/minutes/hours/etc. since YYYY-MM-DD:HH:MM:SS/etc
    # and converts the variable to seconds since epoch
    units = time_var.units
    dates = num2date(time_var[:], units=units)
    unix_times = list()
    for date in dates:
        unix_time = calendar.timegm(date.timetuple())
        unix_times.append(unix_time)
    return unix_times


def netcdf_to_tplot(filenames, time ='', prefix='', suffix='', plot=False, merge=False):
    '''
    This function will automatically create tplot variables from CDF files.

    Parameters:
        filenames : str/list of str
            The file names and full paths of netCDF files.
        time: str
            The name of the netCDF file's time variable.
        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.
        plot: bool
            The data is plotted immediately after being generated.  All tplot
            variables generated from this function will be on the same plot.
            By default, a plot is not created.
        merge: bool
            If True, then data from different netCDF files will be merged into
            a single pytplot variable.

    Returns:
        List of tplot variables created.

    Examples:
        >>> #Create tplot variables from a GOES netCDF file
        >>> import pytplot
        >>> file = "/Users/user_name/goes_files/g15_epead_a16ew_1m_20171201_20171231.nc"
        >>> pytplot.netcdf_to_tplot(file, prefix='mvn_')

        >>> #Add a prefix, and plot immediately.
        >>> import pytplot
        >>> file = "/Users/user_name/goes_files/g15_epead_a16ew_1m_20171201_20171231.nc"
        >>> pytplot.netcdf_to_tplot(file, prefix='goes_prefix_', plot=True)

    '''

    from netCDF4 import Dataset

    stored_variables = []

    if isinstance(filenames, str):
        filenames = [filenames]
    elif isinstance(filenames, list):
        filenames = filenames
    else:
        print("Invalid filenames input.")
        #return stored_variables

    for filename in filenames:

        # Read in file
        file = Dataset(filename, "r+")

        # Creating dictionary that will contain variables and their attributes
        vars_and_atts = {}
        for name, variable in file.variables.items():
            vars_and_atts[name] = {}
            for attrname in variable.ncattrs():
                vars_and_atts[name][attrname] = getattr(variable, attrname)

        # Filling in missing values for each variable with np.nan (if values are not already nan)
        # and saving the masked variables to a new dictionary
        masked_vars = {}  # Dictionary containing properly masked variables
        for var in vars_and_atts.keys():
            reg_var = file.variables[var]
            try:
                var_fill_value = vars_and_atts[var]['missing_value']
                if np.isnan(var_fill_value) != True:
                    # We want to force missing values to be nan so that plots don't look strange
                    var_mask = np.ma.masked_where(reg_var == np.float32(var_fill_value), reg_var)
                    var_filled = np.ma.filled(var_mask, np.nan)
                    masked_vars[var] = var_filled
                elif np.isnan(var_fill_value) == True:
                    # missing values are already np.nan, don't need to do anything
                    var_filled = reg_var
                    masked_vars[var] = var_filled
            except:  # continue # Go to next iteration, this variable doesn't have data to mask (probably just a descriptor variable (i.e., 'base_time')
                var_filled = reg_var
                masked_vars[var] = var_filled

        # Most files are from GOES data, which seems to usually have 'time_tag' in them that contain time information.
        # There is an exception filter below that will allow a user to pick a different time variable if time_tag doesn't exist.
        if time != '':
            time_var = file[time]
            unix_times = change_time_to_unix_time(time_var)

        elif time == '':
            time = input('Please enter time variable name. \nVariable list: {l}'.format(l=vars_and_atts.keys()))
            while True:
                if time not in vars_and_atts.keys():
                    # Making sure we input a valid response (i.e., the variable exists in the dataset), and also avoiding
                    # plotting a time variable against time.... because I don't even know what that would mean and uncover.
                    print('Not a valid variable name, please try again.')
                    continue
                elif time in vars_and_atts.keys():
                    time_var = time
                    unix_times = change_time_to_unix_time(time_var)

        for i,var in enumerate(file.variables):
            # Here, we are making sure that the variables are time-based, otherwise we don't want to store them as tplot variables!
            if 'record' in file[var].dimensions[0] or 'time' in file[var].dimensions[0]:
                # Store the data now, as well as merge variables if that's desired

                var_name = prefix + var + suffix
                to_merge = False
                if (var_name in pytplot.data_quants.keys() and (merge == True)):
                    prev_data_quant = pytplot.data_quants[var_name].values
                    prev_time = pytplot.data_quants[var_name].coords['time'].values
                    to_merge = True

                tplot_data = {'x': unix_times, 'y': masked_vars[var]}
                store_data(var_name, tplot_data)
                if var_name not in stored_variables:
                    stored_variables.append(var_name)

                if to_merge == True:
                    cur_data_quant = pytplot.data_quants[var_name].values
                    cur_time_data = pytplot.data_quants[var_name].coords['time'].values
                    merged_data = [prev_data_quant, cur_data_quant]
                    merged_time = [prev_time, cur_time_data]
                    tplot_data = {'x': np.concatenate(merged_time), 'y': np.concatenate(merged_data)}
                    store_data(var_name, tplot_data)

                # If we are interested in seeing a quick plot of the variables, do it
                if plot:
                    tplot(stored_variables)
            else:
                # If the variable isn't time-bound, we're going to look at the next variable
                continue

    return stored_variables




from pysatdata.utils.dailynames import dailynames
from pysatdata.utils.library_functions import testRemoteDir, testFiles
from pysatdata.utils.download import download
from pysatdata.resources.config import openConfigFile
from pysatdata.satellites.read_goes import *
from pysatdata.satellites.read_rbsp import *
from pysatdata.satellites.read_ace import *
from pysatdata.satellites.read_ace_merged import *
from pysatdata.satellites.read_themis import *
from pysatdata.satellites.read_dmsp import *
from pathlib import Path
from loguru import logger as logging
import os
# %%

def load_sat(trange: list=['2013-11-5', '2013-11-6'],
             satellite: str='goes',
             probe: list=[],
             level: str='2',
             instrument: str='magn',
             datatype: str='sectors',
             suffix: str='',
             cadence: str='4sec', # for EMFISIS and ACE mag data
             coord: str='sm', # for EMFISIS mag data
             rel: str='rel04', # for ECT data
             time: str='time', # For GOES satellites
             get_support_data: bool = False,
             varformat=None,
             varnames: list=[],
             downloadonly: bool=False,
             notplot: bool=False,
             no_update: bool=False,
             time_clip: bool=False,
             usePandas: bool = True,
             usePyTplot: bool = False,
             testRemotePath: bool=False,
             searchFilesFirst: bool=False,
             config_file: str = './pysatdata/resources/config_file.json'):


    global remote_path, out_files, pathformat, tvars

    # config_file = openConfigFile(path=config_file).config_file_sat
    config_file = openConfigFile(satellite).config_file_sat
    # set the download data directories
    if os.environ.get('DATA_DIR'):
        config_file[satellite]['local_data_dir'] = os.sep.join([os.environ['DATA_DIR'], f'{satellite}'])

    if os.environ.get(f'{satellite.upper()}_DATA_DIR'):
        config_file[satellite]['local_data_dir'] = os.environ[f'{satellite.upper()}_DATA_DIR']

    # local data path
    if config_file[satellite]['local_data_dir'] != "sat_data/":
        local_path = f"/{config_file[satellite]['local_data_dir']}{ satellite}"
    else:
        local_path = str(Path.home().joinpath(config_file[satellite]['local_data_dir'], satellite))
    logging.info(f'Local Download Path: {local_path}')



    for prb in probe:
        logging.warning("Selecting the sub path key")
        # test remote Path
        if testRemotePath:
            remote_path, subpathKey, filenameKey, datatype = \
                testRemoteDir(config_file, satellite, prb, instrument, level, datatype)
        else:
            remote_path = config_file[satellite]['remote_data_dir']
            subpathKey = 'subpath'
            filenameKey = 'filename'
        logging.info(f'Remotepath: {remote_path}')
        # subpathKey = "altern_subpath"
        # filenameKey = "altern_filename"
        if instrument == "MagEphem":
            remote_path = config_file[satellite]['remote_subpath'][str(prb)][instrument]["secondRemoteDir"]
        # # datatype = 'pitchangle'


        subpathformat = config_file[satellite]['remote_subpath'][str(prb)][instrument][datatype][subpathKey]
        subpathformat = eval(f"f'{subpathformat}'")
        sat_filename = config_file[satellite]['remote_subpath'][str(prb)][instrument][datatype][filenameKey]
        sat_filename = eval(f"f'{sat_filename}'")
        pathformat = f"{subpathformat}{sat_filename}"
        logging.info(f'Remote_file_path: {pathformat}')

        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange)
        out_files = []
        if searchFilesFirst:
            remote_names, files = testFiles(local_path, remote_names)
            if files is not None:
                for file in files:
                    out_files.append(file)
        # download the files
        files = download(remote_file=remote_names, remote_path=remote_path, local_path=local_path, no_download=no_update)
        if files is not None:
            for file in files:
                out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    if satellite in ['goes16_17', 'goes13_15']:

        tvars = readData_goes(out_files, usePyTplot, usePandas, suffix, time)

    if satellite in ['rbsp', 'erg']:
        tvars = readData_rbsp(out_files, usePyTplot, usePandas, suffix, get_support_data,
                              varformat, varnames, notplot)
    if satellite in ['ace', 'omni']:
        if instrument in ['merged']:
            tvars = readData_ace_merged(out_files, usePyTplot, usePandas, suffix, get_support_data,
                              varformat, varnames, notplot)
        else:
            tvars = readData_ace(out_files, usePyTplot, usePandas, suffix, get_support_data,
                              varformat, varnames, notplot)
    if satellite == 'themis':
        tvars = readData_themis(out_files, usePyTplot, usePandas, suffix, get_support_data,
                              varformat, varnames, notplot)
        
    if satellite == 'dmsp':
        tvars = readData_dmsp(out_files, usePyTplot, usePandas, suffix, get_support_data,
                              varformat, varnames, notplot)
    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars

# if satellite == 'ace':
#     remote_path = config_file[satellite]['remote_data_dir']
#     logging.info(f'Remotepath: {remote_path}')
#
#     if instrument == 'fgm':
#         pathformat = f'mag/level_2_cdaweb/mfi_h{level}/%Y/ac_{datatype}_mfi_%Y%m%d_v??.cdf'
#     elif instrument == 'swe':
#         pathformat = 'swepam/level_2_cdaweb/swe_' + datatype + '/%Y/ac_' + datatype + '_swe_%Y%m%d_v??.cdf'
#     elif instrument == 'epm':
#         pathformat = 'epam/level_2_cdaweb/epm_' + datatype + '/%Y/ac_' + datatype + '_epm_%Y%m%d_v??.cdf'
#     elif instrument == 'cris':
#         pathformat = 'cris/level_2_cdaweb/cris_' + datatype + '/%Y/ac_' + datatype + '_cris_%Y%m%d_v??.cdf'
#     elif instrument == 'sis':
#         pathformat = 'sis/level_2_cdaweb/sis_' + datatype + '/%Y/ac_' + datatype + '_sis_%Y%m%d_v??.cdf'
#     elif instrument == 'ule':
#         pathformat = 'uleis/level_2_cdaweb/ule_' + datatype + '/%Y/ac_' + datatype + '_ule_%Y%m%d_v??.cdf'
#     elif instrument == 'sep':
#         pathformat = 'sepica/level_2_cdaweb/sep_' + datatype + '/%Y/ac_' + datatype + '_sep_%Y%m%d_v??.cdf'
#     elif instrument == 'swics':
#         filename_dtype = datatype.split('_')[1] + '_' + datatype.split('_')[0]
#         pathformat = 'swics/level_2_cdaweb/' + datatype + '/%Y/ac_' + filename_dtype + '_%Y%m%d_v??.cdf'
#
# "remote_subpath": {
#       "ace": {
#         "swepam": {
#           "level_2_cdaweb": {
#             "subpath": "rbsp{prb}/{level}/{instrument}/{datatype}/%Y/",
#             "filename": "rbsp-{prb}_{datatype}_{instrument}-{level}_%Y%m%d_v*.cdf",
#           },
#         },
#       },
#     }



from pysatdata.utils.cdf_to_tplot import *

def readData_themis(files, usePyTplot, usePandas,
                  suffix, get_support_data,
                  varformat, varnames, notplot):

    global time_ind, tvars

    tvars = cdf_to_tplot(files, suffix=suffix, get_support_data=get_support_data,
                         varformat=varformat, varnames=varnames, notplot=notplot)

    if usePyTplot:
        return tvars
    if usePandas:
        pass


import os
import re
import warnings
import requests
import logging
import fnmatch
import datetime
import pkg_resources

from pathlib import Path
from shutil import copyfileobj, copy
from tempfile import NamedTemporaryFile
from html.parser import HTMLParser
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


# the following is used to parse the links from an HTML index file
class LinkParser(HTMLParser):
    def handle_starttag(self, tag, attrs):
        if tag == 'a':
            attrs = {k: v for (k, v) in attrs}
            if 'href' in attrs:
                link = attrs['href']
                # kludge to support http://rbspice?.ftecs.com/
                if '/' in link:
                    link = link.split('/')[-1]
                try:
                    self.links.append((link))
                except AttributeError:
                    self.links = [(link)]
    

def download_file(url=None,
                  filename=None,
                  headers={},
                  username=None,
                  password=None,
                  verify=False,
                  session=None,
                  basic_auth=False):
    """
    Download a file and return its local path; this function is primarily meant to be called by the download function
    
    Parameters:
        url: str
            Remote URL to download
        filename: str
            Local file name
        headers: dict
            Dictionary containing the headers to be passed to the requests get call
        username: str
            Username to be used in HTTP authentication
        password: str
            password to be used in HTTP authentication
        verify: bool
            Flag indicating whether to verify the SSL/TLS certificate
        session: requests.Session object
            Requests session object that allows you to persist things like HTTP authentication through multiple calls
    Returns:
        String containing the local file name
    """

    if session is None:
        session = requests.Session()
        retry = Retry(connect=3, backoff_factor=0.5)
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)
    
    if username is not None:
        session.auth = requests.auth.HTTPDigestAuth(username, password)

    # check if the file exists, and if so, set the last modification time in the header
    # this allows you to avoid re-downloading files that haven't changed
    if os.path.exists(filename):
        mod_tm = (datetime.datetime.utcfromtimestamp(os.path.getmtime(filename))).strftime('%a, %d %b %Y %H:%M:%S GMT')
        headers['If-Modified-Since'] = mod_tm

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ResourceWarning)
        if not basic_auth:
            fsrc = session.get(url, stream=True, verify=verify, headers=headers)
        else:
            fsrc = session.get(url, stream=True, verify=verify, headers=headers, auth=(username, password))

    # need to delete the If-Modified-Since header so it's not set in the dictionary in subsequent calls
    if headers.get('If-Modified-Since') is not None:
        del headers['If-Modified-Since']

    # the file hasn't changed
    if fsrc.status_code == 304:
        logging.info('File is current: ' + filename)
        fsrc.close()
        return filename

    # file not found 
    if fsrc.status_code == 404:
        logging.error('Remote file not found: ' + url)
        fsrc.close()
        return None

    # authentication issues
    if fsrc.status_code == 401 or fsrc.status_code == 403:
        logging.error('Unauthorized: ' + url)
        fsrc.close()
        return None

    if fsrc.status_code == 200:
        logging.info('Downloading ' + url + ' to ' + filename)
    else:
        logging.error(fsrc.reason)
        fsrc.close()
        return None

    ftmp = NamedTemporaryFile(delete=False)

    with open(ftmp.name, 'wb') as f:
        copyfileobj(fsrc.raw, f)

    # make sure the directory exists
    if not os.path.exists(os.path.dirname(filename)) and os.path.dirname(filename) != '':
        os.makedirs(os.path.dirname(filename))

    # if the download was successful, copy to data directory
    copy(ftmp.name, filename)

    fsrc.close()
    ftmp.close()
    os.unlink(ftmp.name)  # delete the temporary file
    
    logging.info('Download complete: ' + filename)

    return filename


def download(remote_path='',
             remote_file='',
             local_path='',
             local_file='',
             headers={},
             username=None,
             password=None,
             verify=True,
             session=None,
             no_download=False,
             last_version=False,
             basic_auth=False,
             regex=False):
    """
    Download one or more remote files and return their local paths.
    Parameters:
        remote_path: str
            String consisting of a common URL base for all remote files
        remote_file: str or list of str
            String or string array of URLs to remote files
        local_path: str
            String consisting of a common local path for all local files
        local_file: str or list of str
            String or string array of local destination file names
        headers: dict
            Dictionary containing the headers to be passed to the requests get call
        username: str
            Username to be used in HTTP authentication
        password: str
            Password to be used in HTTP authentication
        basic_auth: bool
            Flag to indicate that the remote server uses basic authentication
            instead of digest authentication
        verify: bool
            Flag indicating whether to verify the SSL/TLS certificate
        session: requests.Session object
            Requests session object that allows you to persist things like HTTP authentication through multiple calls
        no_download: bool
            Flag to not download remote files
        last_version: bool
            Flag to only download the last in file in a lexically sorted 
            list when multiple matches are found using wildcards
        regex: bool
            Flag to allow regular expressions in the file name matching,
            instead of unix style matching
    Returns:
        String list specifying the full local path to all requested files
    """
    local_file_in = local_file

    if isinstance(remote_path, list):
        logging.error('Remote path must be a string')
        return

    if isinstance(local_path, list):
        logging.error('Local path must be a string')
        return

    if local_path == '':
        local_path = str(Path('').resolve())

    if username is not None and password is None:
        logging.error('Username provided without password')
        return

    if session is None:
        session = requests.Session()
        retry = Retry(connect=3, backoff_factor=0.5)
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)

    if username is not None:
        session.auth = requests.auth.HTTPDigestAuth(username, password)

    if headers.get('User-Agent') is None:
        try:
            release_version = pkg_resources.get_distribution("pyspedas").version
        except pkg_resources.DistributionNotFound:
            release_version = 'bleeding edge'
        headers['User-Agent'] = 'pySPEDAS ' + release_version

    out = []
    index_table = {}

    if not isinstance(remote_file, list):
        remote_file = [remote_file]

    urls = [remote_path+rfile for rfile in remote_file]

    for url in urls:
        resp_data = None
        url_file = url[url.rfind("/")+1:]
        url_base = url.replace(url_file, '')

        # automatically use remote_file locally if local_file is not specified
        if local_file_in == '':
            # if remote_file is the entire url then only use the filename
            if remote_path == '':
                local_file = url_file
            else:
                local_file = url.replace(remote_path, '')

                if local_file == '':  # remote_path was the full file name
                    local_file = remote_path[remote_path.rfind("/")+1:]

        filename = os.path.join(local_path, local_file)

        short_path = local_file[:1+local_file.rfind("/")]

        if not no_download:
            # expand the wildcards in the url
            if '?' in url or '*' in url or regex and no_download is False:
                if index_table.get(url_base) is not None:
                    links = index_table[url_base]
                else:
                    logging.info('Downloading remote index: ' + url_base)

                    # we'll need to parse the HTML index file for the file list
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", category=ResourceWarning)
                        try:
                            if not basic_auth:
                                html_index = session.get(url_base, verify=verify, headers=headers)
                            else:
                                html_index = session.get(url_base, verify=verify, headers=headers, auth=(username, password))
                        except requests.exceptions.ConnectionError:
                            continue

                    if html_index.status_code == 404:
                        logging.error('Remote index not found: ' + url_base)
                        continue

                    if html_index.status_code == 401 or html_index.status_code == 403:
                        logging.error('Unauthorized: ' + url_base)
                        continue

                    # grab the links
                    link_parser = LinkParser()
                    link_parser.feed(html_index.text)

                    try:
                        links = link_parser.links
                        index_table[url_base] = links
                    except AttributeError:
                        links = []

                # find the file names that match our string
                if not regex:
                    # note: fnmatch.filter accepts ? (single character) and * (multiple characters)
                    new_links = fnmatch.filter(links, url_file)
                else:
                    reg_expression = re.compile(url_file)
                    new_links = list(filter(reg_expression.match, links))

                if len(new_links) == 0:
                    logging.info("No links matching pattern %s found at remote index %s",url_file,url_base)

                if last_version and len(new_links) > 1:
                    new_links = sorted(new_links)
                    new_links = [new_links[-1]]

                if '?' in remote_path or '*' in remote_path:
                    # the user specified a wild card in the remote_path
                    remote_path = url_base

                # download the files
                for new_link in new_links:
                    resp_data = download(remote_path=remote_path, remote_file=short_path+new_link,
                                         local_path=local_path, username=username, password=password,
                                         verify=verify, headers=headers, session=session, basic_auth=basic_auth)
                    if resp_data is not None:
                        for file in resp_data:
                            out.append(file)
                session.close()
                continue

            resp_data = download_file(url=url, filename=filename, username=username, password=password, verify=verify,
                                      headers=headers, session=session, basic_auth=basic_auth)
        
        if resp_data is not None:
            if not isinstance(resp_data, list):
                resp_data = [resp_data]
            for file in resp_data:
                out.append(file)
        else:
            # download wasn't successful, search for local files
            logging.info('Searching for local files...')
                
            if local_path == '':
                local_path_to_search = str(Path('.').resolve())
            else:
                local_path_to_search = local_path

            for dirpath, dirnames, filenames in os.walk(local_path_to_search):
                local = local_file[local_file.rfind("/")+1:]
                if not regex:
                    matching_files = fnmatch.filter(filenames, local)
                else:
                    reg_expression = re.compile(local)
                    matching_files = list(filter(reg_expression.match, filenames))

                for file in matching_files:
                    out.append(os.path.join(dirpath, file))

            out = sorted(out)

            if last_version:
                if len(out) > 1:
                    out = [out[-1]]

    session.close()
    return out



import logging
import numpy as np

from pyspedas.analysis.tnormalize import tnormalize
from pyspedas.analysis.tcrossp import tcrossp

from pytplot import get_data, store_data

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

def xgse(mag_temp):
    """
    Generates the 'xgse' transformation matrix
    """
    mag_data = get_data(mag_temp)

    # xaxis of this system is X of the gse system. Z is mag field
    x_axis = np.zeros((len(mag_data.times), 3))
    x_axis[:, 0] = 1

    # create orthonormal basis set
    z_basis = tnormalize(mag_temp, return_data=True)
    y_basis = tcrossp(z_basis, x_axis, return_data=True)
    y_basis = tnormalize(y_basis, return_data=True)
    x_basis = tcrossp(y_basis, z_basis, return_data=True)

    return (x_basis, y_basis, z_basis)


def fac_matrix_make(mag_var_name, other_dim='Xgse', pos_var_name=None, newname=None):
    """
    Generates a field aligned coordinate transformation matrix
    from an input B vector array(and sometimes a position vector array)
    then stores it in a tplot variable

    Input
    ----------
        mag_var_name: str
            tplot variable containing the B-field data

    Parameters
    ----------
        other_dim: str, optional
            The second axis for the field aligned coordinate system (default: Xgse)

        pos_var_name: str, optional
            tplot variable containing the spacecraft position data

        newname: str, optional
            Name of the output tplot variable containing the FAC transformation matrices

    Returns
    ----------

        Name of the tplot variable created.

    """

    mag_data = get_data(mag_var_name)

    if mag_data is None:
        logging.error('Error reading tplot variable: ' + mag_var_name)
        return

    if newname is None:
        newname = mag_var_name + '_fac_mat'

    other_dim = other_dim.lower()

    if other_dim == 'xgse':
        basis = xgse(mag_var_name)

    fac_output = np.zeros((len(mag_data.times), 3, 3))
    fac_output[:, 0, :] = basis[0]
    fac_output[:, 1, :] = basis[1]
    fac_output[:, 2, :] = basis[2]

    store_data(newname, data={'x': mag_data.times, 'y': fac_output})

    return newname





from pysatdata.utils.cdf_to_tplot import *
import warnings
import astropy
import pandas as pd

def readData_ace(files, usePyTplot, usePandas,
                  suffix, get_support_data,
                  varformat, varnames, notplot):

    global time_ind, tvars

    with warnings.catch_warnings():
        # for some reason, ACE CDFs throw ERFA warnings (likely while converting
        # times inside astropy); we're ignoring these here
        # see: https://github.com/astropy/astropy/issues/9603
        warnings.simplefilter('ignore', astropy.utils.exceptions.ErfaWarning)
        tvars = cdf_to_tplot(files, suffix=suffix, get_support_data=get_support_data,
                         varformat=varformat, varnames=varnames, notplot=notplot)

    

    if usePyTplot:
        return tvars
    if usePandas:
        import pytz
        import datetime
        components = ["x", "y", "z"]
        out_dict = {}
        for vas in tvars:
            temp_var = pytplot.data_quants[vas].values
            if len(temp_var.shape) > 1:
                for i, c in enumerate(components):
                    out_dict[f"{vas}_{c}"] = list(temp_var[:, i])
            else:
                out_dict[f"{vas}"] = list(temp_var)
        out_vars_df = pd.DataFrame(out_dict)
        for vas in tvars:
            if 'time' in pytplot.data_quants[vas].coords.keys():
                time = pytplot.data_quants[vas].coords['time'].values
                # time = [datetime.datetime.fromtimestamp(i, pytz.timezone("UTC")) for i in time_ind]
                out_vars_df = pd.DataFrame(out_dict, index=time)
            else:
                out_vars_df = pd.DataFrame(out_dict)
        
    return out_vars_df
        

        



import json
import pysatdata
class openConfigFile():
    def __init__(self, satellite):
        self.satellite = satellite
        pathstring = pysatdata.__file__.replace("\\", "/")
        path = "/".join(pathstring.split("/")[:-1])

        path = f"{path}/resources/{self.satellite}/config_file.json"
        print("loading config file from: ", path)
        with open(path, 'r') as f:
            self.config_file_sat = json.load(f)


import pandas as pd
from pysatdata.utils.netcdf2tplot import *

#%%

def readData_goes(files, usePyTplot, usePandas, suffix, time):

    global time_ind, tvars
    for file in files:
        print(f'Reading... {file}')
        tvars = netcdf_to_tplot(file, suffix=suffix, merge=True, time=time)

    if usePyTplot:
        return tvars
    if usePandas:
        import pytz
        import datetime
        components = ["x", "y", "z"]
        out_dict = {}
        for vas in tvars:
            temp_var = pytplot.data_quants[vas].values
            if len(temp_var.shape) > 1:
                for i, c in enumerate(components):
                    out_dict[f"{vas}_{c}"] = list(temp_var[:, i])
            else:
                out_dict[f"{vas}"] = list(temp_var)

        for vas in tvars:
            if 'time' in pytplot.data_quants[vas].coords.keys():
                time_ind = pytplot.data_quants[vas].coords['time'].values
                break
        time = [datetime.datetime.fromtimestamp(i, pytz.timezone("UTC")) for i in time_ind]

        out_vars_df = pd.DataFrame(out_dict, index=time)

        return out_vars_df





"""Print the version number for the current installation."""


def version():
    """
    Print the pysatdata version number.

    Returns
    -------
    None.

    """
    import importlib.metadata
    ver = importlib.metadata.version('pysatdata')
    print("dstdownloader version: " + ver)



# Copyright 2020 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for
# Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import cdflib

# If the user has astropy installed, use the cdflib's CDFAstropy class for time conversion
# (Converting to unix time is much, much faster this way)
try:
    from cdflib.epochs_astropy import CDFAstropy as cdfepoch
except:
    from cdflib.epochs import CDFepoch as cdfepoch

import re
import numpy as np
import xarray as xr
from pytplot.store_data import store_data
from pytplot.tplot import tplot
from pytplot.options import options
import pytplot
import copy


def cdf_to_tplot(filenames, varformat=None, get_support_data=False,
                 get_ignore_data=False, string_encoding='ascii',
                 prefix='', suffix='', plot=False, merge=False,
                 center_measurement=False, notplot=False, varnames=[]):
    """
    This function will automatically create tplot variables from CDF files.  In general, the files should be
    ISTP compliant for this importer to work.  Each variable is read into a new tplot variable (a.k.a an xarray DataArray),
    and all associated file/variable metadata is read into the attrs dictionary.

    .. note::
        Variables must have an attribute named "VAR_TYPE". If the attribute entry
        is "data" (or "support_data"), then they will be added as tplot variables.
        Additionally, data variables should have attributes named "DEPEND_TIME" or
        "DEPEND_0" that describes which variable is x axis.  If the data is 2D,
        then an attribute "DEPEND_1" must describe which variable contains the
        secondary axis.

    Parameters:
        filenames : str/list of str
            The file names and full paths of CDF files.
        varformat : str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.
        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".
        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.
        plot: bool
            The data is plotted immediately after being generated.  All tplot
            variables generated from this function will be on the same plot.
        merge: bool
            If True, then data from different cdf files will be merged into
            a single pytplot variable.
        get_ignore_data: bool
            Data with an attribute "VAR_TYPE" with a value of "ignore_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".
        center_measurement: bool
            If True, the CDF epoch variables are time-shifted to the middle
            of the accumulation interval by their DELTA_PLUS_VAR and
            DELTA_MINUS_VAR variable attributes
        notplot: bool
            If True, then data are returned in a hash table instead of
            being stored in tplot variables (useful for debugging, and
            access to multi-dimensional data products)
        varnames: str or list of str
            Load these variables only. If [] or ['*'], then load everything.

    Returns:
        List of tplot variables created (unless notplot keyword is used).
    """

    stored_variables = []
    epoch_cache = {}
    output_table = {}
    metadata = {}

    if not isinstance(varnames, list):
        varnames = [varnames]
    
    if len(varnames) > 0:
        if '*' in varnames:
            varnames = []

    # pytplot.data_quants = {}
    if isinstance(filenames, str):
        filenames = [filenames]
    elif isinstance(filenames, list):
        filenames = filenames
    else:
        print("Invalid filenames input.")
        return stored_variables

    var_type = ['data']
    if varformat is None:
        varformat = ".*"
    if get_support_data:
        var_type.append('support_data')
    if get_ignore_data:
        var_type.append('ignore_data')

        
    varformat = varformat.replace("*", ".*")
    var_regex = re.compile(varformat)
    filenames.sort()
    for filename in filenames:
        cdf_file = cdflib.CDF(filename)
        cdf_file.string_encoding = string_encoding
        cdf_info = cdf_file.cdf_info()
        all_cdf_variables = cdf_info['rVariables'] + cdf_info['zVariables']
        # User defined variables.
        if len(varnames) > 0:
            load_cdf_variables = [value for value in varnames if value in all_cdf_variables]
        else:
            load_cdf_variables = all_cdf_variables

        try:
            gatt = cdf_file.globalattsget()
        except:
            gatt={}

        for var in load_cdf_variables:
            if not re.match(var_regex, var):
                continue
            var_atts = cdf_file.varattsget(var)

            if 'VAR_TYPE' not in var_atts:
                continue

            if var_atts['VAR_TYPE'] in var_type:
                var_atts = cdf_file.varattsget(var)
                var_properties = cdf_file.varinq(var)

                # Find data name and if it is already in stored variables
                if 'TPLOT_NAME' in var_atts:
                    var_name = prefix + var_atts['TPLOT_NAME'] + suffix
                else:
                    var_name = prefix + var + suffix

                if "DEPEND_TIME" in var_atts:
                    x_axis_var = var_atts["DEPEND_TIME"]
                elif "DEPEND_0" in var_atts:
                    x_axis_var = var_atts["DEPEND_0"]
                else:
                    # non-record varying variables (NRVs)
                    # added by egrimes, 13Jan2021
                    # here we assume if there isn't a DEPEND_TIME or DEPEND_0, there are no other depends
                    try:
                        ydata = cdf_file.varget(var)
                    except:
                        continue

                    if ydata is None:
                        continue

                    # since NRVs don't vary with time, they shouldn't vary across files
                    output_table[var_name] = {'y': ydata}

                    continue

                data_type_description \
                    = cdf_file.varinq(x_axis_var)['Data_Type_Description']

                if epoch_cache.get(filename+x_axis_var) is None:
                    delta_plus_var = 0.0
                    delta_minus_var = 0.0
                    delta_time = 0.0

                    # Skip variables with ValueErrors. 
                    try:
                        xdata = cdf_file.varget(x_axis_var)
                        epoch_var_atts = cdf_file.varattsget(x_axis_var)
                    except ValueError:
                        print("Value error for variable: " + var_name)
                        continue

                    # check for DELTA_PLUS_VAR/DELTA_MINUS_VAR attributes
                    if center_measurement:
                        if 'DELTA_PLUS_VAR' in epoch_var_atts:
                            delta_plus_var = cdf_file.varget(epoch_var_atts['DELTA_PLUS_VAR'])
                            delta_plus_var_att = cdf_file.varattsget(epoch_var_atts['DELTA_PLUS_VAR'])

                            # check if a conversion to seconds is required
                            if 'SI_CONVERSION' in delta_plus_var_att:
                                si_conv = delta_plus_var_att['SI_CONVERSION']
                                delta_plus_var = delta_plus_var.astype(float)*np.float(si_conv.split('>')[0])
                            elif 'SI_CONV' in delta_plus_var_att:
                                si_conv = delta_plus_var_att['SI_CONV']
                                delta_plus_var = delta_plus_var.astype(float)*np.float(si_conv.split('>')[0])

                        if 'DELTA_MINUS_VAR' in epoch_var_atts:
                            delta_minus_var = cdf_file.varget(epoch_var_atts['DELTA_MINUS_VAR'])
                            delta_minus_var_att = cdf_file.varattsget(epoch_var_atts['DELTA_MINUS_VAR'])

                            # check if a conversion to seconds is required
                            if 'SI_CONVERSION' in delta_minus_var_att:
                                si_conv = delta_minus_var_att['SI_CONVERSION']
                                delta_minus_var = delta_minus_var.astype(float)*np.float(si_conv.split('>')[0])
                            elif 'SI_CONV' in delta_minus_var_att:
                                si_conv = delta_minus_var_att['SI_CONV']
                                delta_minus_var = delta_minus_var.astype(float)*np.float(si_conv.split('>')[0])

                        # sometimes these are specified as arrays
                        if isinstance(delta_plus_var, np.ndarray) and isinstance(delta_minus_var, np.ndarray):
                            delta_time = (delta_plus_var-delta_minus_var)/2.0
                        else: # and sometimes constants
                            if delta_plus_var != 0.0 or delta_minus_var != 0.0:
                                delta_time = (delta_plus_var-delta_minus_var)/2.0

                if epoch_cache.get(filename + x_axis_var) is None:
                    if ('CDF_TIME' in data_type_description) or \
                            ('CDF_EPOCH' in data_type_description):
                        xdata = cdfepoch.unixtime(xdata)
                        epoch_cache[filename+x_axis_var] = np.array(xdata)+delta_time
                else:
                    xdata = epoch_cache[filename + x_axis_var]

                try:
                    ydata = cdf_file.varget(var)
                except:
                    continue

                if ydata is None:
                    continue
                if "FILLVAL" in var_atts:
                    if (var_properties['Data_Type_Description'] ==
                            'CDF_FLOAT' or
                            var_properties['Data_Type_Description'] ==
                            'CDF_REAL4' or
                            var_properties['Data_Type_Description'] ==
                            'CDF_DOUBLE' or
                            var_properties['Data_Type_Description'] ==
                            'CDF_REAL8'):

                        if ydata[ydata == var_atts["FILLVAL"]].size != 0:
                            ydata[ydata == var_atts["FILLVAL"]] = np.nan

                tplot_data = {'x': xdata, 'y': ydata}


                # Data may depend on other data in the CDF.
                depend_1 = None
                depend_2 = None
                depend_3 = None
                if "DEPEND_1" in var_atts:
                    if var_atts["DEPEND_1"] in all_cdf_variables:
                        depend_1 = np.array(cdf_file.varget(var_atts["DEPEND_1"]))
                        # Ignore the depend types if they are strings
                        if depend_1.dtype.type is np.str_:
                            depend_1 = None
                if "DEPEND_2" in var_atts:
                    if var_atts["DEPEND_2"] in all_cdf_variables:
                        depend_2 = np.array(cdf_file.varget(var_atts["DEPEND_2"]))
                        # Ignore the depend types if they are strings
                        if depend_2.dtype.type is np.str_:
                            depend_2 = None
                if "DEPEND_3" in var_atts:
                    if var_atts["DEPEND_3"] in all_cdf_variables:
                        depend_3 = np.array(cdf_file.varget(var_atts["DEPEND_3"]))
                        # Ignore the depend types if they are strings
                        if depend_3.dtype.type is np.str_:
                            depend_3 = None

                nontime_varying_depends = []

                if depend_1 is not None and depend_2 is not None and depend_3 is not None:
                    tplot_data['v1'] = depend_1
                    tplot_data['v2'] = depend_2
                    tplot_data['v3'] = depend_3

                    if len(depend_1.shape) == 1:
                        nontime_varying_depends.append('v1')
                    if len(depend_2.shape) == 1:
                        nontime_varying_depends.append('v2')
                    if len(depend_3.shape) == 1:
                        nontime_varying_depends.append('v3')

                elif depend_1 is not None and depend_2 is not None:
                    tplot_data['v1'] = depend_1
                    tplot_data['v2'] = depend_2
                    if len(depend_1.shape) == 1:
                        nontime_varying_depends.append('v1')
                    if len(depend_2.shape) == 1:
                        nontime_varying_depends.append('v2')
                elif depend_1 is not None:
                    tplot_data['v'] = depend_1
                    if len(depend_1.shape) == 1:
                        nontime_varying_depends.append('v')
                elif depend_2 is not None:
                    tplot_data['v'] = depend_2
                    if len(depend_2.shape) == 1:
                        nontime_varying_depends.append('v')

                metadata[var_name] = {'display_type': var_atts.get("DISPLAY_TYPE", "time_series"),
                                      'scale_type': var_atts.get("SCALE_TYP", "linear"),
                                      'var_attrs': var_atts, 'file_name': filename, 'global_attrs': gatt}

                # Check if the variable already exists in the for loop output
                if var_name not in output_table:
                    output_table[var_name] = tplot_data
                else:
                    # If it does, loop though the existing variable's x,y,v,v2,v3,etc
                    var_data = output_table[var_name]
                    for output_var in var_data:
                        if output_var not in nontime_varying_depends:
                            if np.asarray(tplot_data[output_var]).ndim == 0 and np.equal(tplot_data[output_var], None):
                                # If there is nothing in the new variable, then pass
                                pass
                            elif np.asarray(var_data[output_var]).ndim == 0 and np.equal(var_data[output_var], None):
                                # If there is nothing in the old variable, then replace
                                var_data[output_var] = tplot_data[output_var]
                            else:  # If they both have something, then concatenate
                                var_data[output_var] = np.concatenate((var_data[output_var], tplot_data[output_var]))

    if notplot:
        return output_table

    for var_name in output_table.keys():
        to_merge = False
        if var_name in pytplot.data_quants.keys() and merge:
            prev_data_quant = pytplot.data_quants[var_name]
            to_merge = True

        try:
            attr_dict = {}
            if metadata.get(var_name) is not None:
                attr_dict["CDF"] = {}
                attr_dict["CDF"]["VATT"] = metadata[var_name]['var_attrs']
                attr_dict["CDF"]["GATT"] = metadata[var_name]['global_attrs']
                attr_dict["CDF"]["FILENAME"] = metadata[var_name]['file_name']

            # extract the coordinate system, if available
            # try:
            #     vatt_keys = list(attr_dict["CDF"]["VATT"].keys())
            #     vatt_lower = [k.lower() for k in vatt_keys]
            #     if 'coordinate_system' in vatt_lower:
            #         attr_dict['data_att'] = {'coord_sys': attr_dict["CDF"]["VATT"][vatt_keys[vatt_lower.index('coordinate_system')]]}
            # except (Exception) as e:
            #     print(e)
            #
            if 'v' in list(output_table[var_name].keys()):
                if len(output_table[var_name]['v']) != len(output_table[var_name]['x']):
                    yy = output_table[var_name]['y']
                    vv = output_table[var_name]['v']
                    if len(vv.shape) > 1:
                        aas = np.zeros((len(yy), vv.shape[1]))
                        for i in range(len(yy)):
                            aas[i,:] = vv[0,:]
                        output_table[var_name]['v'] = aas
                    else:
                        aas = np.zeros((len(yy), len(vv)))
                        for i in range(len(yy)):
                            aas[i,:] = vv
                        output_table[var_name]['v'] = aas

            store_data(var_name, data=output_table[var_name], attr_dict=attr_dict)
        except ValueError:
            continue

        if var_name not in stored_variables:
            stored_variables.append(var_name)

        if metadata.get(var_name) is not None:
            if metadata[var_name]['display_type'] == "spectrogram":
                options(var_name, 'spec', 1)
            if metadata[var_name]['scale_type'] == 'log':
                options(var_name, 'ylog', 1)
            # Gather up all options in the variable attribute section, toss them into options and see what sticks
            options(var_name, opt_dict=metadata[var_name]['var_attrs'])

        if to_merge is True:
            cur_data_quant = pytplot.data_quants[var_name]
            plot_options = copy.deepcopy(pytplot.data_quants[var_name].attrs)
            pytplot.data_quants[var_name] = xr.concat([prev_data_quant, cur_data_quant], dim='time').sortby('time')
            pytplot.data_quants[var_name].attrs = plot_options

    if notplot:
        return output_table

    if plot:
        tplot(stored_variables)

    return stored_variables



import datetime
import numpy as np
import pandas as pd
from scipy import interpolate as interp
from pysatdata.utils.library_functions import fill_nan


def interpolateFluxRbsp(enSignal, lValues, timeArray, resolution_L=0.025):
    """
    enSignal: electron flux at one energy level (1D array)
    lValues: L-shell or l-star time series
    timeArray: datetime array 

    """

    L_inerp = fill_nan(lValues)

    tHour0 = timeArray[0].hour
    tHour1 = timeArray[-1].hour
    tMin0 = timeArray[0].minute
    tMin1 = timeArray[-1].minute
    tSec0 = timeArray[0].second
    tSec1 = timeArray[-1].second
    dad0 = int(timeArray[0].strftime('%j')) + ((tHour0 / 24) + (tMin0 / (60*24)) + (tSec0 / (60*60*24)))
    dad1 = int(timeArray[-1].strftime('%j')) + ((tHour1 / 24) + (tMin1 / (60*24)) + (tSec1 / (60*60*24)))

    time = np.linspace(dad0 , dad1, len(L_inerp))
    p = np.matrix.transpose(np.asmatrix([time, L_inerp]))
    z = enSignal

    xtime, xL = np.meshgrid(np.arange(min(time), max(time), 0.01), np.arange(min(L_inerp), max(L_inerp),resolution_L))

    # datetime x array
    xax = [datetime.datetime(timeArray[0].year, 1, 1, 0, 0) + datetime.timedelta(t - 1) for t in xtime[0,:]]

    # creade Y axis array (L values)
    yax = np.arange(min(L_inerp), max(L_inerp),resolution_L)

    # interpolate the data indo the grid
    flux = interp.griddata(p, z, (xtime, xL), method='linear')

    # Elimitade nan values
    flux[flux == 0.0] = 'NaN'
    maskflux = np.ma.masked_where(np.isnan(flux),flux, copy=True)

    return xax, yax, maskflux



"""A setuptools based setup module."""
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open

    
setup(
    name='pysatdata',
    version='1.0.11',
    description='Python Space Physics Satellite Data Analysis Toolkit',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/zemarchezi/pySatData',
    author='Jose Paulo Marchezi',
    author_email='jose.marchezi@inpe.br, jpmarchezi@gmail.com',
    license='MIT',
    classifiers=['Development Status :: 2 - Pre-Alpha',
                 'Intended Audience :: Science/Research',
                 'Topic :: Scientific/Engineering',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3',
                 ],
    keywords='satellite space data tools',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=['requests', 'pyspedas', 'pytplot-mpl-temp>=2.1.17',
                      'cdflib>=0.3.20', 'cdasws>=1.7.24', 'netCDF4',
                      'pywavelets', 'astropy','pyqtgraph', 'loguru',
                      'aacgmv2>=2.6.2', 'matplotlib', 'netCDF4',
                      'pandas', 'pytz', 'scipy', 'tqdm',
                      'xarray','urllib3'],
    python_requires='>=3.6',
    include_package_data=True,
)


import os
import matplotlib
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.ticker import *
import datetime
from pysatdata.utils.library_functions import replace_at_index1, format_func, cutFlux_lshell
import matplotlib.colors as colors
import matplotlib.cm as cm
from pathlib import Path
from loguru import logger as logging

def plot_classicFluxSWparams(**kwargs):

    for key, val in kwargs.items():
        globals()[key] = val


    out_figDir = str(Path.home().joinpath('Pictures', plotDir))

    if not os.path.exists(out_figDir) and os.path.dirname(out_figDir) != '':
        os.makedirs(out_figDir)

    xLim=[datetime.datetime.strptime(trangeXlim[0], '%Y-%m-%d'), datetime.datetime.strptime(trangeXlim[1], '%Y-%m-%d')]

    matplotlib.rc('font', size=fontsize)
    plt.close('all')
    plt.ioff()
    figprops = dict(figsize=(16,18), dpi=120)
    fig = plt.figure(**figprops)

    energChanel = f'{specEnergy[fluxEnergyChanel]}'
    if level == 2:
        om = ' -- OMNI directional'
    else:
        om = ''

    titlePanel1 = f'RBSP {probe}: ECT/REPT (L{level}) Electron Flux density, Energy {energChanel} MeV{om}'

    if figureIdentify != None:
        figureFilename = f'{figureIdentify}_RBSP{probe}_ECT-REPT_L{level}_Electron-Flux-density-Energy_{energChanel}' \
                         f'_MeV{trangeXlim[0]}_{trangeXlim[1]}.png'
    else:
        figureFilename = f'RBSP{probe}_ECT-REPT_L{level}_Electron-Flux-density-Energy_{energChanel}' \
                         f'_MeV{trangeXlim[0]}_{trangeXlim[1]}.png'


    ax = plt.axes([0.055, 0.79, 0.817, 0.18])
    plt.subplots_adjust(left=0.06, right=0.875, bottom=0.07, top=0.95)
    if interpolatedFlux == True:
        lc = ax.pcolormesh(xax, yax, maskflux, norm=colors.LogNorm(vmin=vmin, vmax = vmax), cmap='jet')
    else:
        lc = ax.scatter(time_dt_rept,l_probe, 8, flux_rept_spec[:,fluxEnergyChanel], norm=colors.LogNorm(vmin=vmin, vmax= vmax), cmap='jet')
    ax.text(0.05, 0.9, '(a)', horizontalalignment='center',verticalalignment='center',
            fontsize=18, transform=ax.transAxes)
    ax.set_title(titlePanel1)
    ax.set_ylabel(f'{lshellOrlStar}')
    # ax.set_xlabel('Time (UTC)', fontsize=15)
    ax.set_xlim(xLim[0], xLim[1])
    ax.set_xticklabels([])
    ax.xaxis.set_minor_locator(AutoMinorLocator())
    ax.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    ax.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    cbar_coord = replace_at_index1(make_axes_locatable(ax).get_position(), [0,2], [0.9, 0.01])
    cbar_ax = fig.add_axes(cbar_coord)
    cbar = fig.colorbar(lc, cax=cbar_ax)
    cbar.set_label('$[cm^{-2}s^{-1}sr^{-1}MeV^{-1}]$')

    bx = plt.axes([0.055, 0.59, 0.817, 0.18])
    cor = cm.tab10(np.linspace(0, 1, len(energyRange)))
    for il in range(len(energyRange)):
        cut_date, cutF = cutFlux_lshell(flux_rept_spec, cutLshell, il, l_probe, time_dt_rept)
        bx.semilogy(cut_date, cutF,  '-*', color=cor[il], label="{:1.1f} MeV".format(specEnergy[energyRange[il]]) )
        bx.legend(loc='upper right', bbox_to_anchor=(1.15, 1.02))
    bx.set_xlim(xLim[0], xLim[1])
    bx.text(0.05, 0.9, '(b)', horizontalalignment='center', verticalalignment='center',
            fontsize=18, transform=bx.transAxes)
    bx.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    bx.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    bx.get_xaxis().set_ticks_position('both')
    bx.get_yaxis().set_ticks_position('both')
    bx.set_xticklabels([])
    bx.set_title(f'Cut in flux at ${lshellOrlStar} = {cutLshell}$')
    bx.set_ylabel('$Flux / [cm^{-2}s^{-1}sr^{-1}MeV^{-1}]$')
    bx.grid()
    #
    # # Flow Speed
    cx = plt.axes([0.055, 0.48, 0.817, 0.1])
    cx.plot(time_dt_swe, flow_speed, '-', color='b')
    cx.get_xaxis().set_ticks_position('both')
    cx.get_yaxis().set_ticks_position('both')
    cx.text(0.05, 0.9, '(c)', horizontalalignment='center', verticalalignment='center',
            fontsize=18, transform=cx.transAxes)
    # cx.legend(loc='upper right', bbox_to_anchor=(1.12, 1.02))
    cx.set_ylabel('$V_{sw} / [Km s^{-1}]$')
    cx.set_xticklabels([])
    cx.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    cx.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    cx.set_xlim(xLim[0], xLim[1])
    cx.grid()
    #
    # ## density
    dx = plt.axes([0.055, 0.37, 0.817, 0.1])
    dx.plot(time_dt_swe, nP, '-', color='b')
    dx.get_xaxis().set_ticks_position('both')
    dx.get_yaxis().set_ticks_position('both')
    dx.text(0.05, 0.9, '(d)', horizontalalignment='center', verticalalignment='center',
            fontsize=18, transform=dx.transAxes)
    dx.legend(loc='upper right', bbox_to_anchor=(1.11, 1.02))
    dx.set_ylabel('$Np / [cm^{-3}]$')
    dx.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    dx.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    dx.set_xlim(xLim[0], xLim[1])
    dx.set_xticklabels([])
    dx.grid()
    #
    #
    # ## imf
    ex = plt.axes([0.055, 0.26, 0.817, 0.1])
    ex.plot(time_dt_swe, bgse_x, '-', color='b', label="Bx_gse")
    ex.plot(time_dt_swe, bgse_y, '-', color='r', label="By_gse")
    ex.get_xaxis().set_ticks_position('both')
    ex.get_yaxis().set_ticks_position('both')
    ex.text(0.05, 0.9, '(e)', horizontalalignment='center', verticalalignment='center',
            fontsize=18, transform=ex.transAxes)
    ex.legend(loc='upper right', bbox_to_anchor=(1.14, 1.02))
    ex.set_ylabel('$IMF / [nT]$')
    ex.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    ex.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    ex.set_xlim(xLim[0], xLim[1])
    ex.set_xticklabels([])
    ex.grid()
    #
    # ## Bz
    fx = plt.axes([0.055, 0.15, 0.817, 0.1])
    fx.plot(time_dt_swe, bgse_z, '-', color='b', label="Bz_gse")
    fx.plot(time_dt_swe, b_total, '-', color='r', label="B total")
    fx.get_xaxis().set_ticks_position('both')
    fx.get_yaxis().set_ticks_position('both')
    fx.text(0.05, 0.9, '(f)', horizontalalignment='center', verticalalignment='center',
            fontsize=18, transform=fx.transAxes)
    fx.legend(loc='upper right', bbox_to_anchor=(1.14, 1.02))
    fx.set_ylabel('$IMF / [nT]$')
    fx.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    fx.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    fx.set_xlim(xLim[0], xLim[1])
    fx.set_xticklabels([])
    fx.grid()
    #
    #
    # ## AE index
    gx = plt.axes([0.055, 0.04, 0.817, 0.1])
    gx.plot(time_dt_swe, bgse_z, '-', color='b', label="AE Index")
    gx.get_xaxis().set_ticks_position('both')
    gx.get_yaxis().set_ticks_position('both')
    gx.text(0.05, 0.9, '(g)', horizontalalignment='center', verticalalignment='center',
            fontsize=18, transform=gx.transAxes)
    gx.legend(loc='upper right', bbox_to_anchor=(1.145, 1.02))
    gx.set_ylabel('$/ [nT]$')
    gx.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    gx.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    gx.set_xlim(xLim[0], xLim[1])
    gx.xaxis.set_major_formatter(plt.FuncFormatter(format_func))
    gx.grid()

    logging.info(f'saving figure at: {out_figDir}/{figureFilename}')
    plt.savefig(f'{out_figDir}/{figureFilename}')


#%%

def plot_classicSWparams(**kwargs):

    for key, val in kwargs.items():
        globals()[key] = val


    out_figDir = str(Path.home().joinpath('Pictures', plotDir))

    if not os.path.exists(out_figDir) and os.path.dirname(out_figDir) != '':
        os.makedirs(out_figDir)

    xLim=[datetime.datetime.strptime(trangeXlim[0], '%Y-%m-%d'), datetime.datetime.strptime(trangeXlim[1], '%Y-%m-%d')]

    matplotlib.rc('font', size=fontsize)
    plt.close('all')
    plt.ioff()
    figprops = dict(figsize=(16,18), dpi=120)
    fig = plt.figure(**figprops)


    if level == 2:
        om = ' -- OMNI directional'
    else:
        om = ''

    figureFilename = f'RBSP{probe}_ECT-REPT_L{level}_' \
                     f'_MeV{trangeXlim[0]}_{trangeXlim[1]}.png'

    bx = plt.axes([0.055, 0.59, 0.817, 0.1])
    bx.plot(time_dt_swe, flow_speed, '-', color='b')
    bx.set_xlim(xLim[0], xLim[1])
    bx.text(0.05, 0.9, '(b)', horizontalalignment='center', verticalalignment='center',
            fontsize=18, transform=bx.transAxes)
    bx.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    bx.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    bx.get_xaxis().set_ticks_position('both')
    bx.get_yaxis().set_ticks_position('both')
    bx.set_xticklabels([])
    bx.set_ylabel('$Vp [Km s^{-1}]$')
    bx.grid()
    #
    # # Flow Speed
    cx = plt.axes([0.055, 0.48, 0.817, 0.1])
    cx.plot(time_dt_swe, flow_speed, '-', color='b', label="Vsw")
    cx.get_xaxis().set_ticks_position('both')
    cx.get_yaxis().set_ticks_position('both')
    cx.text(0.05, 0.9, '(c)', horizontalalignment='center', verticalalignment='center',
            fontsize=18, transform=cx.transAxes)
    cx.legend(loc='upper right', bbox_to_anchor=(1.12, 1.02))
    cx.set_ylabel('$Vp [Km s^{-1}]$')
    cx.set_xticklabels([])
    cx.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    cx.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    cx.set_xlim(xLim[0], xLim[1])
    cx.grid()
    #
    # ## density
    dx = plt.axes([0.055, 0.37, 0.817, 0.1])
    dx.plot(time_dt_swe, nP, '-', color='b', label="Np")
    dx.get_xaxis().set_ticks_position('both')
    dx.get_yaxis().set_ticks_position('both')
    dx.text(0.05, 0.9, '(d)', horizontalalignment='center', verticalalignment='center',
            fontsize=18, transform=dx.transAxes)
    dx.legend(loc='upper right', bbox_to_anchor=(1.11, 1.02))
    dx.set_ylabel('$\\# / [cm^{-3}]$')
    dx.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    dx.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    dx.set_xlim(xLim[0], xLim[1])
    dx.set_xticklabels([])
    dx.grid()
    #
    #
    # ## imf
    ex = plt.axes([0.055, 0.26, 0.817, 0.1])
    ex.plot(time_dt_swe, bgse_x, '-', color='b', label="Bx_gse")
    ex.plot(time_dt_swe, bgse_y, '-', color='r', label="By_gse")
    ex.get_xaxis().set_ticks_position('both')
    ex.get_yaxis().set_ticks_position('both')
    ex.text(0.05, 0.9, '(e)', horizontalalignment='center', verticalalignment='center',
            fontsize=18, transform=ex.transAxes)
    ex.legend(loc='upper right', bbox_to_anchor=(1.14, 1.02))
    ex.set_ylabel('$\\# / [nT]$')
    ex.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    ex.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    ex.set_xlim(xLim[0], xLim[1])
    ex.set_xticklabels([])
    ex.grid()
    #
    # ## Bz
    fx = plt.axes([0.055, 0.15, 0.817, 0.1])
    fx.plot(time_dt_swe, bgse_z, '-', color='b', label="Bz_gse")
    fx.plot(time_dt_swe, b_total, '-', color='r', label="B total")
    fx.get_xaxis().set_ticks_position('both')
    fx.get_yaxis().set_ticks_position('both')
    fx.text(0.05, 0.9, '(f)', horizontalalignment='center', verticalalignment='center',
            fontsize=18, transform=fx.transAxes)
    fx.legend(loc='upper right', bbox_to_anchor=(1.14, 1.02))
    fx.set_ylabel('$\\# / [nT]$')
    fx.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    fx.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    fx.set_xlim(xLim[0], xLim[1])
    fx.set_xticklabels([])
    fx.grid()
    #
    #
    # ## AE index
    gx = plt.axes([0.055, 0.04, 0.817, 0.1])
    gx.plot(time_dt_swe, ae_index, '-', color='b', label="AE Index")
    gx.get_xaxis().set_ticks_position('both')
    gx.get_yaxis().set_ticks_position('both')
    gx.text(0.05, 0.9, '(g)', horizontalalignment='center', verticalalignment='center',
            fontsize=18, transform=gx.transAxes)
    gx.legend(loc='upper right', bbox_to_anchor=(1.145, 1.02))
    gx.set_ylabel('$\\# / [nT]$')
    gx.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    gx.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    gx.set_xlim(xLim[0], xLim[1])
    gx.xaxis.set_major_formatter(plt.FuncFormatter(format_func))
    gx.grid()

    logging.info(f'saving figure at: {out_figDir}/{figureFilename}')
    plt.savefig(f'{out_figDir}/{figureFilename}')


#%%

def plot_FluxvsL(**kwargs):

    for key, val in kwargs.items():
        globals()[key] = val


    out_figDir = str(Path.home().joinpath('Pictures', plotDir))

    if not os.path.exists(out_figDir) and os.path.dirname(out_figDir) != '':
        os.makedirs(out_figDir)

    xLim=[datetime.datetime.strptime(trangeXlim[0], '%Y-%m-%d'), datetime.datetime.strptime(trangeXlim[1], '%Y-%m-%d')]

    matplotlib.rc('font', size=fontsize)
    plt.close('all')
    plt.ioff()
    figprops = dict(figsize=(16,4), dpi=120)
    fig = plt.figure(**figprops)

    energChanel = f'{specEnergy[fluxEnergyChanel]}'
    if level == 2:
        om = ' -- OMNI directional'
    else:
        om = ''

    titlePanel1 = f'RBSP {probe}: ECT/REPT (L{level}) Electron Flux density, Energy {energChanel} MeV{om}'

    if figureIdentify != None:
        figureFilename = f'{figureIdentify}_RBSP{probe}_ECT-REPT_L{level}_Electron-Flux-density-Energy_{energChanel}' \
                         f'_MeV{trangeXlim[0]}_{trangeXlim[1]}.png'
    else:
        figureFilename = f'RBSP{probe}_ECT-REPT_L{level}_Electron-Flux-density-Energy_{energChanel}' \
                         f'_MeV{trangeXlim[0]}_{trangeXlim[1]}.png'


    ax = plt.axes([0.055, 0.79, 0.817, 0.18])
    plt.subplots_adjust(left=0.06, right=0.875, bottom=0.07, top=0.95)
    if interpolatedFlux == True:
        lc = ax.pcolormesh(xax, yax, maskflux, norm=colors.LogNorm(vmin=vmin, vmax = vmax), cmap='jet')
    else:
        for ff in fluxAB:
            lc = ax.scatter(time_dt_rept,l_probe, 8, ff, norm=colors.LogNorm(vmin=vmin, vmax= vmax), cmap='jet')
    ax.text(0.05, 0.9, '(a)', horizontalalignment='center',verticalalignment='center',
            fontsize=18, transform=ax.transAxes)
    ax.set_title(titlePanel1)
    ax.set_ylabel(f'{lshellOrlStar}')
    # ax.set_xlabel('Time (UTC)', fontsize=15)
    ax.set_xlim(xLim[0], xLim[1])
    ax.set_xticklabels([])
    ax.xaxis.set_minor_locator(AutoMinorLocator())
    ax.tick_params(direction='in', length=10, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='major')
    ax.tick_params(direction='in', length=7, width=0.7, colors='k',
                   grid_color='k', grid_alpha=0.5, which='minor')
    cbar_coord = replace_at_index1(make_axes_locatable(ax).get_position(), [0,2], [0.9, 0.01])
    cbar_ax = fig.add_axes(cbar_coord)
    cbar = fig.colorbar(lc, cax=cbar_ax)
    cbar.set_label('$[cm^{-2}s^{-1}sr^{-1}MeV^{-1}]$')

    logging.info(f'saving figure at: {out_figDir}/{figureFilename}')
    plt.savefig(f'{out_figDir}/{figureFilename}')


#%%
from pysatdata.utils.cdf_to_tplot import *
import pandas as pd
import numpy as np
import datetime
import gzip

#%%
def readData_ace_merged(files, usePyTplot, usePandas,
                  suffix, get_support_data,
                  varformat, varnames, notplot):

    global time_ind, tvars
    columns = ["Year", "Decimal Day", "Hour", "Minute",
           "X_GSE", "Y_GSE", "Z_GSE", "B_Magnitude", 
           "Bx_GSE", "By_GSE", "Bz_GSE", "No_of_Vectors",
           "Quality_flag", "Proton_density", "Temperature",
           "HeH_Ratio", "Flow_speed", "Vx", "Vy", "Vz"]
    al_data = []
    for ff in files:
        da = []
        with gzip.open(ff,'rt') as f:
            for line in f:
                da.append(np.asarray(list(filter(None, line.split("\n")[0].split(" ")))).astype(float))

        da = np.asarray(da)
        timeDt = [datetime.datetime.strptime(f'{int(da[i][0])}{int(da[i][1])} {int(da[i][2])}:{int(da[i][3])}', '%Y%j %H:%M') for i in range(len(da))]

        mergedData = pd.DataFrame(da, index=timeDt, columns=columns)
        
        al_data.append(mergedData)
    
    mergedAce = pd.concat(al_data)

    if usePyTplot:
        print("Not implemented yet, returning pandas Dataframe")
        return mergedAce
    if usePandas:
        return mergedAce
        
        

        

# %%
